import importlib.metadata
from math import pi, tau

VERSION = importlib.metadata.version("ghedesigner")

PI = pi
DEG_TO_RAD = pi / 180.0
RAD_TO_DEG = 180.0 / pi
PI_OVER_2 = pi / 2.0
TWO_PI = tau
HRS_IN_DAY = 24
SEC_IN_HR = 3600
MONTHS_IN_YEAR = 12
